//Tabella delle associazioni di un NameServer;
package Dns;


import java.util.*;
import java.io.*;

public class BindingsTable implements Serializable{
	//tabella dei bindings:
	protected transient Hashtable table;
	
	//serializzazione////////////////////////////
	private void writeObject(ObjectOutputStream out) throws IOException {
		//l'hashtable non viene serializzata
		//dimensione della tabella:
		out.writeInt(table.size());
		Enumeration enum=(Enumeration)table.keys();
		//elementi della tabella:
		while(enum.hasMoreElements()){
			DNSName key=(DNSName)enum.nextElement();
			ArrayList list=(ArrayList)table.get(key);
			out.writeObject(key);
			out.writeInt(list.size());
			for(int i=0;i<list.size();i++)
				out.writeObject(list.get(i));
		}//while
	}//writeObject
	
	private void readObject(ObjectInputStream in) throws IOException,ClassNotFoundException{
		table=new Hashtable();
		for(int count=in.readInt();count>0;count--){
			DNSName key=(DNSName)in.readObject();
			ArrayList list=new ArrayList();
			int size=in.readInt();
			for(int i=0;i<size;i++)
				list.add(in.readObject());
			table.put(key,list);
		}
	}//readObject
	
	/////////////////////////////////////////////
	
	public BindingsTable(){
		table=new Hashtable();
	}//costruttore
	
	//inserimento di una nuova entry:
	public synchronized void put(Binding entry){
		ArrayList list;
		DNSName key=entry.getName();
		if (table.containsKey(key)){
			list =(ArrayList)table.get(key);
		} else {
			list=new ArrayList();
		}
		//controllo che non ci sia:
		int i=0;
		for(;i<list.size();i++){
				if(list.get(i).equals(entry.getValue()))break;
		}
		if(i==list.size())list.add(entry.getValue());
		table.put(key,list);
	}//put
	
	//rimozione di una entry:
	public synchronized void remove(DNSName name)throws BindNotFoundException{
		if (table.containsKey(name)){
			table.remove(name);
			System.out.println("Eliminazione effettuata");
		}else {
			System.out.println("Nome non presente");
			throw new BindNotFoundException();
		}
	}//remove
	
	//ricerca di un nome:
	public synchronized Value[] get(DNSName name) throws BindNotFoundException{
		ArrayList list=(ArrayList)table.get(name);
		//System.out.println("Cerco: "+name);
		//System.out.println("Contenuto "+table);
		//System.out.println("Trovato "+list);
		if (list==null) throw new BindNotFoundException();
		else {
			Object[] obj= list.toArray();
			Value[] value=new Value[obj.length];
			for(int i=0;i<value.length;i++){
				value[i]=(Value)obj[i];
			}
			return value;
		}
	}//get

	public boolean isEmpty(){
		return table.isEmpty();
	}//isEmpty
	
	public boolean contains(DNSName name){
		return table.containsKey(name);
	}//contains
	
	public String toString(){
		return table.toString();
	}
	
}//class BindingsTable
