package Dns;

//////////////////////////////////////////////////////////
// classe per il caching di DNSBindings  
// - ogni Binding ha un tempo di vita (in sec.) che vale
//   dal momento della ricezione
// - allo scadere del TTL il binding perde di validit
//
// Si pu prevedere un controllo che avviene ad intervalli regolari e
// accettabili (ad es. ogni n minuti) e che elimina le entry scadute
//
// Oppure associare una lista di coppie valore-scadenza ordinata
// e innescare un time-out ogni volta.
// ad es. al secondo 0 ricevo b1 con TTL 10:
// lista (b1,10) innsesco un time-out a 10
// al secondo 2 ricevo b2 con ttl 5:
// lista (b2,7)-(b1,10) reinnesco il t-o a 5
// al secondo 4 ricevo b3 con ttl 5
// lista (b2,7)-(b3,9)-(b1,10)
// al secondo 7 scatta il t-o; elimino la prima entry, 
//		re-imposto il t-0 a 9-7=2
// ecc. ecc.
// 
// Non mi vengono in mente altre soluzioni poich la tabella
// deve rimanere con hashing sui nomi per motivi di efficienza della ricerca
//
// La seconda soluzione  migliore dal punto di vista della consistenza
// ma  necessario mantenere una lista che come struttura dati  della stessa dimensione
// della tabella poich oltre al nome bisogna mantenere un riferimento a quale degli indirizzi relativi
// deve essere eliminato. Inoltre prevede una attivit continua da parete del thread incaricato
// di eliminare le entry scadute per cui la serializzazione dei thread potrebbe bloccare i richiedenti
// molto spesso in attesa che il monitor si sblocchi. 
//
// Per il momento sacrificheremo la consistenza per la faclit e l'efficienza della realizzazione
// del primo metodo, anche perch questo servizio non richiede tempi di caching brevi ma molto 
// lunghi (dell'ordine di grandezza delle ore o addirittura dei giorni).

import javax.swing.Timer;
import java.net.InetAddress;
import java.awt.event.*;
import java.util.*;

public class Cache extends BindingsTable implements ActionListener {
	//il timer che si risveglia a intervalli regolari
	protected Timer timeout;
	//costante che indica ogni quanti secondi si effettua il controllo
	//tutti i TTL vengono arrotondati a un multiplo di questo valore.
	protected int checktime=300; //5 minuti di default
	//protected int checktime=20; //prova 20 sec
	

	
	
	//costruttore che mantiene l'intervallo di default
	public Cache(){
		super();
		timeout=new Timer(checktime*1000,this);
		timeout.start();
	}//costruttore
	
	//costruttore che specifica l'intervallo 
	public Cache(int checktime){
		super();
		this.checktime=checktime;
		timeout=new Timer(checktime*1000,this);
		timeout.start();
	}//costruttore
	
	//controllo e eliminazione entry scadute dalla tabella:
	public void actionPerformed(ActionEvent e){
		//System.out.println("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
		System.out.println("CACHE: controllo,cache prima= "+table);
		//per ogni entry della tabella diminuisce il TTL di uno
		//quando si arriva a zero l'indirizzo viene eliminato
		for (Enumeration enum = table.keys();enum.hasMoreElements();) {
    	DNSName key=(DNSName)enum.nextElement();
			ArrayList list=(ArrayList)table.get(key);
				for(int i=0;i<list.size();i++){
					int TTL=((CacheValue)list.get(i)).getTimeToLive();
					if((--TTL)==0){
						list.remove(i--);
						if(list.isEmpty())table.remove(key);
					}else ((CacheValue)list.get(i)).setTimeToLive(TTL);
				} 
     }
		 System.out.println("CACHE: controllo,cache dopo= "+table);
		 //System.out.println("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
	}//actionPerformed
	
	//overriding dei metodi di accesso alla tabella 
	//necessario inserire ancche il TTL corrispondente:
	//inserimento di una nuova entry:
	public synchronized void put(Binding entry,int TTL){
		ArrayList list;
		DNSName key=entry.getName();
		if (table.containsKey(key)){
			list =(ArrayList)table.get(key);
		} else {
			list=new ArrayList();
		}
		//calcolo l'istante di scadenza:
		//durata del TTL in clock del timer:
		//int since=getNoTimeout()+(TTL/checktime)+1;
		int since=(TTL/checktime)+1;
		//se il TTL non  divisibile si prende l'istante precedente
		//DA TOGLIERE
		if(since==0){
			System.err.println("CACHE: L'intervallo di controllo  troppo lungo!");
			since=1;
		}
		//controllo se non  gia presente
		boolean found=false;
		for(int i=0;i<list.size();i++){
			CacheValue value=(CacheValue)list.get(i);
			if ( (value.getAddress().equals(entry.getValue().getAddress()))&& (value.getPort()==entry.getValue().getPort())){
				//entry presente aggiorno il TTL se necessario
				if(value.getTimeToLive()<since)value.setTimeToLive(since);
				found=true;
			}
		}
		if(!found)list.add(new CacheValue(entry.getValue(),since));
		table.put(key,list);
		//System.out.println("inserimento, cache= "+ table);
	}//put
	
	public synchronized Value[] get(DNSName name) {
		ArrayList list=(ArrayList)table.get(name);
		//System.out.println("CACHE: Cerco: "+name);
		//System.out.println("Trovato "+list);
		if (list==null) return null;
		else {
			Object[] obj= list.toArray();
			Value[] value=new Value[obj.length];
			for(int i=0;i<value.length;i++){
				CacheValue val=(CacheValue)obj[i];
				value[i]=new Value(val.getAddress(),val.getPort()); 
			}
			return value;
		}
	}//get
	

	
	
	//il TTL restituito per un nome  dato dal minore dei TTTL
	//attualmente presenti per i suoi valori
	public synchronized int getTimeToLive(DNSName name){
		ArrayList list=(ArrayList)table.get(name);
		//System.out.println("CACHE: Cerco TTL di: "+name);
		//System.out.println("CACHE: Trovato "+list);
		if (list==null) return -1;
		else {
			int min=((CacheValue)list.get(0)).getTimeToLive();
			for(int i=1;i<list.size();i++){
				int ttl=((CacheValue)list.get(i)).getTimeToLive(); 
				if (ttl<min)min=ttl;
			}
			//System.out.println("CACHE: MIN="+ min);
			return min;
		}
	}//getTimeToLive
	
	
	/*
	//incrementa il numero dei timeout scattati
	// synchronized per evitare letture e 
	//scritture di notimeout contemporane da parte di thread diversi
	synchronized void incNoTimeout(){
		notimeout++;
	}//incNoTimeOut
	
	synchronized int getNoTimeout(){
		return notimeout; 
	}//incNoTimeOut
	*/

}//class Cache


