package Dns;

// Astrazione di un nome;
// Si pu considerare costituito da:
//	-un identificativo	(il nome utente,o identificativo di un servizio, 
//   nel nostro caso NS o AS);
//  -il dominio di appartenenza (realizzato da una lista di nomi atomici 
//   che, partendo dalla root, arrivano al dom. finale)
//	 Pu essere eventualmente nullo per spazi di nomi "flat".

import java.io.*;
import javax.naming.InvalidNameException;


public class DNSName implements Serializable{
	//identificativo 
	protected Type id;					
	
	//lista dei domini componenti, cammino a partire dalla root;
	//il primo elemento  sempre la root.
	protected Domain dom;	
	
	//serializzazione dell'oggetto
	private void writeObject(ObjectOutputStream out) throws IOException {
		
		out.writeObject(id);
		out.writeObject(dom);
	}
	
	private void readObject(ObjectInputStream in) throws IOException,ClassNotFoundException{
		id=(Type)in.readObject();
		dom=(Domain)in.readObject();
	}
	
	
	//costruttore
	public DNSName(){
		id=null;
		dom=null;
	}
	
	//costruttore con parsing del nome
	public DNSName(String s) throws InvalidNameException{
		NameParser np=new Parser(s);
		dom=np.getDomain();
		id=np.getType();
	}//costruttore

	//costruttore con parsing del solo cammino
	public DNSName(Type id,String s) throws InvalidNameException{
		this.id=id;
		NameParser np=new Parser(s);
		dom=np.getDomain();
	}//costruttore
	
	//costruttore con campi
	public DNSName(Type id,Domain dom) throws InvalidNameException{
		this.id=id;
		this.dom=dom;
	}//costruttore
	
	public Type getType(){
		return id;
	}//getType
	
	public Domain getDomain(){
		return dom;
	}//getType

	public String toString(){
		return "Name="+id+dom;
	}//toString
	
	public int hashCode(){
	 return id.hashCode()+dom.hashCode();
	}//hashCode

	//necessario per l'hashing:
	public boolean equals(Object o){
		if(o instanceof DNSName){	
			if ((this.id.equals(((DNSName)o).getType()))&&(this.dom.equals(((DNSName)o).getDomain())))
				return true;
		}
		return false;
	}//equals
	
}//class DNSName
