package Dns;

//Classe che generalizza il nome del dominio
//Sintassi dei nomi di dominio:
//Dominio  ::= root | alfa '.' Dominio
//root ::=""
//
//I nomi di dominio sono assoluti
//per cui dovrebbe esserci il punto finale sempre ("b.a.")
//dove non presente ("b.a") si intende sottointeso
//Con alfa si intendono tutti i caratteri 

import java.util.*;
import javax.naming.*;
import java.io.*;

public class Domain  implements Serializable,Name{
	//il dominio vero e proprio:
	transient CompoundName name;
	
	//La root  indicata con:
	public static final String rootName="";
	
	//serializzazione dell'oggetto
	private void writeObject(ObjectOutputStream out) throws IOException {
		out.writeObject(toString());
	}
	
	private void readObject(ObjectInputStream in) throws IOException,ClassNotFoundException{
		String dom=(String)in.readObject();
		try{
			name=new CompoundName(dom,getSyntax());
		}	catch(InvalidNameException e){throw new ClassNotFoundException(e.getMessage());}
	}
	//serializzazione
	
	
	public static Domain getRootName()throws InvalidNameException{
		if (rootName.equals(""))
			return new Domain(".");
		else return new Domain(rootName);
	}//getRootName

	//costruttore senza argomenti per serializzazione
	public Domain(){
		name=null;
	}//costruttore
	
	public Domain(String n) throws InvalidNameException{
		name=new CompoundName(n,getSyntax());
		check();
	}//costruttore
	
	public Domain(CompoundName n) throws InvalidNameException{
		name=n;
		check();
	}//costruttore

	
	protected void check() throws InvalidNameException{
		//controllo della validit del nome
		//non ci deve essere un componente con nome uguale alla root
		//se non al primo posto
		//se non c' al primo posto si aggiunge
		if (name.size()==0)name.add("");
		if(!name.get(0).equals(rootName))name.add(0,rootName);
		for(int i=1;i<name.size();i++){
			if(name.get(i).equals(rootName))
				throw new InvalidNameException("Can't use root name.");
		}//for
	}//check

	//metodo che definisce la sintassi del nome;
	//eventuali sottoclassi possono fare overriding per cambiarla;
	public static Properties getSyntax(){
		Properties syntax=new	Properties();
		syntax.setProperty("jndi.syntax.direction","right_to_left");
		syntax.setProperty("jndi.syntax.separator",".");
		syntax.setProperty("jndi.syntax.ignorecase","true");
		return syntax;
	}//getSyntax
	
	public boolean isRoot(){
		//se  impossibile creare il nome di root ritorna sempre false
		try{
			return (name.compareTo(getRootName().name)==0);
		} catch(InvalidNameException e){
			return false;
		}
	}//isRoot

	public Object clone() {
		try{
			return new Domain((CompoundName)name.clone());
		}catch(InvalidNameException e){
			return null;
		}
	}
	
	public Name addAll(int posn,Name n)throws InvalidNameException{
		return name.addAll(posn,n);
	}//addAll
	
	// Adds a single component at a specified position within this name.
	public Name add(int posn, String comp) throws InvalidNameException{
 		return name.add(posn,comp);
	}
  

 	public Name add(String comp) throws InvalidNameException{
		return name.add(comp);
	}
          
 	public Name addAll(Name suffix) throws InvalidNameException{
		if(suffix instanceof Domain){
			name.addAll(((Domain)suffix).name);
			return this;
		}	else if (suffix instanceof CompoundName){
			name.addAll(suffix);
			return this;
		} else {
			throw new InvalidNameException("argument must be a Domain");
		}
	}
	
	public int compareTo(Object obj){
		if(obj==null || !(obj instanceof Domain))throw new ClassCastException();
		else return this.name.compareTo(((Domain)obj).name);
	}
	
 	public boolean endsWith(Name n){
		if(n instanceof Domain){
			return name.endsWith(((Domain)n).name);
		}	else return false;
	}
	
 	public String get(int posn){ 
 		return name.get(posn);
	}
 
	public Enumeration getAll(){
		return name.getAll();
	}
 	
	public Name getPrefix(int posn) {
		try{
		CompoundName c=(CompoundName)name.getPrefix(posn);
		return new Domain(c);
		}catch(InvalidNameException e){
			return null;
		}
	}
 
	public Name getSuffix(int posn){
		//non ritorna un Domain ma un
		//CompoundName poich un Domain deve sempre 
		//cominciare con la Root
		CompoundName c=(CompoundName)name.getSuffix(posn);
		return c;
	}
 
	public boolean isEmpty(){
		return name.isEmpty();
	}
          
 	public Object remove(int posn) throws InvalidNameException{ 
 		return name.remove(posn);
	}
 
	public int size(){
 		return name.size();
	}
          
 	public boolean startsWith(Name n){
		if(n instanceof Domain){
			return name.startsWith(((Domain)n).name);
		} else return false;
	}
	
	public boolean equals(Object obj){
		if(obj instanceof Domain){
			return name.equals(((Domain)obj).name);
		}else return false;
	}

	public int hashCode(){
		return name.hashCode();	
	}

	public String toString(){
		return name.toString();
	}

}//class Domain
