package Dns;

import java.net.*;
import javax.naming.InvalidNameException;
import javax.naming.CompoundName;
import java.io.InterruptedIOException;
import Dns.Msg.*;
import Msg.PacketMsg;

public class IterQueryResolver extends Thread {
	protected QueryMsg query;          			//nome da risolvere
	protected InetAddress senderaddr;	 			//indirizzo a cui inviare la risposta
	protected int senderport;          			//porta     a cui inviare la risposta
	protected NameServer parent;					 	//il NS che ha attivato il thread
	protected boolean localsender;			 		//il richiedente  locale?
	protected int selectaddr=0;				 			//indice dell'ultimo indirizzo scelto 
	protected static final int MaxIter=15; 	//numero massimo di iterazioni nel processo iterativo
	protected static final int MaxRetry=3;	//numero massimo di tentativi nel reinvio di un messaggio
	protected String tname;
	
	////////////////////////////////////////////////////////////////////////////////
	// __________________																													//
	//|									 |																												//
	//|Metodo iterativo  |																												//
	//|__________________|																												//
	//Bisogna distinguere le richieste da dominio locale(A)(Resolver o 						//
	//NS non registrati) da quelle non locali(B).																	//
	//																																						//
	//Nel primo caso infatti  questo NS che si deve far carico del processo 			//
	//iterativo inviando le successive query																			//
	//																																						//
	//Risoluzione di una query:																										//
	//-------------------------																										//																																				
	//		1- se il nome richiesto  locale  o di un sottodominio non delegato			//
	//			si risponde immediatamente																						//
	//		altrimenti:																															//
	//		2- Controllo nella cache della presenza del nome, o di una "scorciatoia"//
  //			per raggiungerlo(ad es. sottodomini di un sottodominio delegato)			//
	//																																						//
	//(A) Parte la ricerca iterativa (A):																					//
	//																																						//
	//		PRIMO PASSO																															//
	//		1- se il nome richiesto  di un sottodominio delegato										//
	//			(l'indirizzo del NS deve essere presente!)														//
	//			si invia il la query al NS corrispondente altrimenti l'indirizzo 			//
	//			richiesto 	errato.																										//
	//		altrimenti:																															//
	//*-->2- se c' una radice comune tra il dominio richiesto e il dominio del NS//
	//				,si predilige il metodo ITERATIVO VERSO L'ALTO.											//
	//        Inoltre si pu effettuare un controllo sui "passi" da compiere			//
	//		3- altrimenti si invia la richiesta alla root.													//
	//				(L'indirizzo della Root DEVE dempre essere presente)								//
	//																																						//
	//		PASSI SUCCESSIVI Finch non si riceve un AnswerMsg:											//
	//		Alla ricezione di un ReferTo si invia la query all'indirizzo corr.			//
	//																																						//
	//   	CONCLUSIONE invio risposta al richiedente 															//
	//																																						//
	//(B) Controllo e risposta (B):																								//
 	// 		Poich il nome richiesto non  locale si cerca se  un sottodominio del //
	//			si invia il ReferTo, altrimenti (RICERCA IT. VERSO L'ALTO) si invia il//
 	//		ReferTo	del NS superiore se presente (ma solo dopo aver controllato che //
	//		non ci sia un	un prefisso in comune tra il proprio dom. e quello 				//
	//		richiesto.In questo caso vorrebbe dire che c' un errore-->si rischiano //
	//		CICLI infiniti.																													//
	////////////////////////////////////////////////////////////////////////////////
	
	public IterQueryResolver(QueryMsg query,InetAddress senderaddr,int senderport,NameServer parent){
		this.query=query;
		this.senderaddr=senderaddr; 
		this.senderport=senderport;
		this.parent=parent;
		if(query instanceof NameQueryMsg)
			this.localsender=((NameQueryMsg)query).isLocal();
		tname=getName();
	}//costruttore

	
	public void run(){
		if(query instanceof DomainQueryMsg){
			parent.writeOut(tname+": Richiesta del dominio.");
			int tsquery=query.getTimestamp();
			AnswerMsg amsg=new DomainAnswerMsg(parent.getDomain(),tsquery);
			sendMessage(amsg,senderaddr,senderport);			
		} else if(query instanceof NameQueryMsg){
			Domain querydomain=((NameQueryMsg)query).getQuery().getDomain();
			parent.writeOut(tname+": Richiesta per "+((NameQueryMsg)query).getQuery());
			Domain domain=parent.getDomain();
			//parent.writeOut("Dominio del NS: "+domain);
			//timestamp della richiesta:
			int tsquery=query.getTimestamp();
			//parent.writeOut("timestamp della richiesta: "+tsquery);
			
			
			
			//Controllo localit del nome richiesto:
			//controllo se  locale o se  usn sottodominio non delegato: il problema  che potrebbbe essere delegato uno qualsiasi dei 
			//dei sottodomini che compongono il nome, ad es:
			//se siamo in b.a e il nome  e.d.c.b.a potrebbe esere delegato c.b.a ma anche esserlo solo d.c.b.a
			//Devo controllarli tutti? per forza.
			//Mantenendo una lista dei sottodomini la ricerca  pi veloce di quella nella tabella.
			//La ricerca potrebbe risultare molto lunga in caso di domini errati.
					
			//Ho quindi aggiunto alla dnstable una lista dei domini non delegati e di quelli delegati:
			//parent.writeOut("querydomain.compareTo(domain)==0:"+(querydomain.compareTo(domain)==0));
			//parent.writeOut("parent.isSubdomain(querydomain):"+(parent.isSubdomain(querydomain)));
			//parent.writeOut("parent.isSubdomainDel(querydomain):"+(parent.isSubdomainDel(querydomain)));
			if(querydomain.compareTo(domain)==0 || 
			(parent.isSubdomain(querydomain) && !(parent.isSubdomainDel(querydomain)))){
				//Questo  il NS competente invio comunque la risposta:
				parent.writeOut(tname+": NameServer competente. ");
				AnswerMsg amsg;
				try{	
					Value[] answ=parent.lookup(((NameQueryMsg)query).getQuery());
					amsg=new BindFoundMsg(((NameQueryMsg)query).getQuery(),answ,parent.getTimeToLive(),tsquery);
					//parent.writeOut("(*) "+getName()+": Risposta trovata");
				} catch (BindNotFoundException e){
					//il nome richiesto non esiste
					amsg=new BindNotFoundMsg(((NameQueryMsg)query).getQuery(),tsquery);
					//parent.writeOut("(*) "+getName()+":Risposta non trovata");
				}			
				sendMessage(amsg,senderaddr,senderport);
				parent.writeOut(tname+": Inviata risposta "+amsg);
			} else {			
				//il nome richiesto non  locale:
				/////////////////////////////////////
				//ricerca nella cache;             //
				/////////////////////////////////////
				
				Value[] refer=null;
				DNSName cachefound=null;
				SignedMsg recvmsg=null;
				int ttl=0;

				cachefound=parent.cacheSearch(((NameQueryMsg)query).getQuery());
				//parent.writeOut("cachfound "+cachefound);
				if(!(cachefound==null)){
					refer=parent.cacheLookup(cachefound);
					ttl=parent.getCacheTimeToLive(cachefound);
					if(cachefound.equals(((NameQueryMsg)query).getQuery())){
						parent.writeOut(tname+": Trovato risultato in Cache:"+cachefound+" in "+refer);
						recvmsg=new BindFoundMsg(((NameQueryMsg)query).getQuery(),refer,ttl,tsquery);
					}else{
						parent.writeOut(tname+": Trovato refer in Cache a:"+cachefound+" in "+refer);
					}
				}
				
				//try{System.in.read();}catch(Exception e){}
				/////////////////////////////////////////////////////////////////////
				//controllo la localit del sender
				//parent.writeOut("Questo non e' il NS competente");
				if(localsender){
					parent.writeOut(tname+": Ricerca iterativa");
				//(A)RICERCA ITERATIVA:
					//PRIMO PASSO:solo se non ho trovato niente nella cache
					Domain querytodomain=null;
					if(cachefound==null){
							//controllo che il dominio richiesto sia un sottodominio
							if(querydomain.startsWith(domain)){
							//Controllo qual' il sottodominio delegato corrispondente:
								//parent.writeOut(tname+": Cerco il sottodominio delegato competente:");
								querytodomain=findSubdomainDel(querydomain);
								parent.writeOut(tname+": Sottodominio delegato competente :"+querytodomain);
							}	else 
							//E' conveniente utilizzare la ricerca verso l'alto?
							if(iterUpSearch(querydomain)){
								//Ricerca iterativa verso l'alto
								querytodomain=parent.getParentDomain();
							} else{
								//richiesta alla root
								try{
									querytodomain=Domain.getRootName();
								} catch (InvalidNameException e){
									System.err.println(tname+": Errore fatale nella gestione del NS:"+
									" Impossibile generare il nome di Root"+e);
									this.stop();
								}//try
							}//if
					}//if not cachefound
					else parent.writeOut(tname+": Invio risposta trovata in cache.");
					DatagramSocket sock=null;
					//Value[] refer=null;
					try{
						//Creazione di una socket datagram per il ciclo iterativo
						//su  una porta qualsiasi disponibile
						sock=new DatagramSocket();
						//tempo di attesa della risposta per la socket
						sock.setSoTimeout(NameServer.timeOut *1000);
						if(refer==null)refer=parent.lookup(new NSName(querytodomain)); 
					} catch(Exception ine){
						System.err.println(tname+": Errore fatale nel thread.");
						if(sock!=null)sock.close();
						this.stop();
					}
					//PASSI SUCCESSIVI DEL CICLO ITERATIVO:
		
					boolean done=false;
					Value addr;
					//trasformo la query locale in query non loc.
					int ts=parent.stamp();
					//parent.writeOut("il mio timestamp  "+ts);
					query=new NameQueryMsg(((NameQueryMsg)query).getQuery(),false,ts);
					int itercount=0;
					//se ho trovato il risultato in cache non entra nel while:
					while((!(recvmsg instanceof AnswerMsg))&&(itercount++<MaxIter)){
						//parent.writeOut(" (recvmsg instanceof AnswerMsg)"+(recvmsg instanceof AnswerMsg));
						//parent.writeOut("_________PASSO N"+itercount+"____________");
						//invio della richiesta corrente e ricezione risposta:
						//BISOGNA INSERIRE UN MAX AI TENTATIVI PER NON BLOCCARSI
						//AD ES PER PROBLEMI DI RETE
						selectaddr=0;
						done=false;
						int retrycount=0;
						while(!done && (retrycount++<MaxRetry)){
							//scelta tra gli indirizzi disponibili
							addr=selectAddr(refer);
							//aggiorno il timestamp;
							ts=parent.stamp();
							//parent.writeOut("il mio nuovo timestamp  "+ts);
							query.setTimestamp(ts);
							parent.writeOut(tname+": passo n"+itercount+"invio richiesta a "+addr+".");
							//preparazione e invio pacchetto
							try{
								DatagramPacket packet=query.writeToPacket(addr.getAddress(),addr.getPort());
								sock.send(packet);
								//parent.writeOut("Richiesta inviata");
							} catch(Exception e){
								System.err.println(tname+": Impossibile inviare a " +addr.getAddress().getHostName()+
								":"+addr.getPort()+" la richiesta "+query.toString()+": "+ e.getMessage());
							}
							//attesa risposta con timeout
							try{
								DatagramPacket recvpacket=new DatagramPacket(new byte[PacketMsg.dimPacket],PacketMsg.dimPacket);
								sock.receive(recvpacket);
								recvmsg=(SignedMsg)SignedMsg.readFromPacket(recvpacket);
								parent.writeOut(tname+": passo n"+itercount+",ricevuto "+recvmsg+".");
								//controllo timestamp:
								if ((!(recvmsg instanceof TimestampedMsg)) || (((TimestampedMsg)recvmsg).getTimestamp()!=ts)){
									System.err.println("Errore :Messaggio replicato o Timestamp non valido");
									this.stop();
								}
								done=true;
							}catch(InterruptedIOException ie){
								done=false;
							}catch(Exception e){
								System.err.println("Errore :"+e.getMessage());
								done=false;
							}//try
						}//while
						//nel caso in cui non siamo riusciti a mandare il messaggio
						//terminiamo la ricerca iterativa
						if(!done)itercount=MaxIter;
						//la risposta  stata ricevuta o  scattato un timeout
						//caching della risposta
						if(recvmsg instanceof CachableMsg){
							refer=((CachableMsg)recvmsg).getValue();
							DNSName recvname;
							if(recvmsg instanceof ReferToMsg)recvname=((ReferToMsg)recvmsg).getName();
								else recvname=((NameQueryMsg)query).getQuery();
							for(int i=0;i<refer.length;i++){
								parent.cache(new Binding(recvname,refer[i]),((CachableMsg)recvmsg).getTimeToLive());
							}
						}//if
					}//while
					//Ottenuta un risposta alla query
					//reinvio al richiedente
					try{
						if(!(itercount<MaxIter))
							recvmsg=new BindNotFoundMsg(((NameQueryMsg)query).getQuery(),tsquery);
						parent.writeOut(tname+": Invio risposta "+recvmsg);
						((TimestampedMsg)recvmsg).setTimestamp(tsquery);
						DatagramPacket packet=recvmsg.writeToPacket(senderaddr,senderport);
						//politica at-most once:
						sock.send(packet);
						sock.close();
					}catch (Exception e) {
						System.err.println(tname+": Impossibile inviare risposta a "+senderaddr.getHostName()+
						":"+senderport+" per la richiesta "+query.toString()+": "+ e.getMessage());
						sock.close();
					}
				} else{
				parent.writeOut(tname+": Controllo e Risposta.");
				//(B)CONTROLLO E RISPOSTA
				PacketMsg amsg=null;
				if(cachefound==null){
							//controllo se  un sottodominio
								if(querydomain.startsWith(domain)){
								//Il nome ha un prefisso in comune potrebbe essere un sottodominio:
								//Controllo qual' il sottodominio delegato corrispondente:
									//parent.writeOut("Prefisso in comune");
									Domain deldom=findSubdomainDel(querydomain);
									parent.writeOut(tname+": Trovato sottodominio delegato "+deldom+".");
									if(deldom!=null)
										try{	
											Value[] answ=parent.lookup(new NSName(deldom));
											amsg=new ReferToMsg(new NSName(deldom),answ,parent.getTimeToLive(),tsquery);
											parent.writeOut(tname+": Inviato Refer");
										} catch (BindNotFoundException bnfe){
											//il nome richiesto non esiste, MA DEVE ESISTERE!
											System.err.println(tname+": Errore nella gestione del NS:"+bnfe.getMessage());
											amsg=new BindNotFoundMsg(((NameQueryMsg)query).getQuery(),tsquery);
										} catch (InvalidNameException ine){
											System.err.println(tname+": Errore fatale nella gestione del NS:"+ine.getMessage());
											this.stop();
										}//try
									else amsg=new BindNotFoundMsg(((NameQueryMsg)query).getQuery(),tsquery);
								}	else {
								//non  un sottodominio, ricerca it. verso l'alto
									if(!domain.isRoot()){
										//ricerca iterativa verso l'alto
										parent.writeOut(tname+": Ricerca iterativa verso l'alto");
										Domain parentdomain=parent.getParentDomain();
										try{	
											Value[] answ=parent.lookup(new NSName(parentdomain));
											amsg=new ReferToMsg(new NSName(parentdomain),answ,parent.getTimeToLive(),tsquery);
										} catch (BindNotFoundException bnfe){
											//il nome richiesto non esiste, MA DEVE ESISTERE!
											System.err.println(tname+": Errore nella gestione del NS:"+bnfe.getMessage());
											amsg=new BindNotFoundMsg(((NameQueryMsg)query).getQuery(),tsquery);
										} catch (InvalidNameException ine){
											System.err.println(tname+": Errore fatale nella gestione del NS:"+ine.getMessage());
											this.stop();
										}
									}else {
										//se il dominio  root allore DEVE essere un sottodominio:
										System.err.println(tname+": Errore nella gestione del NS: sottodominio non trovato!");
										amsg=new BindNotFoundMsg(((NameQueryMsg)query).getQuery(),tsquery);
										
									}
								}//if  un sottodominio
					}	else {
					 	if(cachefound.equals(((NameQueryMsg)query).getQuery()))
							amsg=new BindFoundMsg(((NameQueryMsg)query).getQuery(),refer,ttl,tsquery);
						else
							amsg=new ReferToMsg(cachefound,refer,ttl,tsquery);
					}//if not found cache
					//invio della risposta:
					sendMessage(amsg,senderaddr,senderport);
				}//if sender  locale
			}//if dominio richiesto   locale 
		}
	}//run
	
	//procedura che dato un nome di dominio trova il 
	//primo sottodominio delegato a cui fare riferimento
	
	protected Domain findSubdomainDel(Domain querydomain){
		//parent.writeOut("Ricerca sott. competente per "+querydomain);
		Domain domain=parent.getDomain();
		//parent.writeOut("siamo in "+domain);
		if(querydomain.startsWith(domain)){
			//parent.writeOut("domain.size()= "+domain.size());
			CompoundName subs=(CompoundName)querydomain.getSuffix(domain.size());
			//parent.writeOut("Nomi da aggiungere=" +subs);
			//for(int j=0;j<subs.size();j++){
			//parent.writeOut("["+j+"]= " +subs.get(j));
			//parent.writeOut(subs.getPrefix(j+1));
			//}
			for(int i=0;i<subs.size();i++){
				try{
					Domain d=(Domain)domain.clone();
					Domain searchdeldom=(Domain)d.addAll(subs.getPrefix(i+1));
					//parent.writeOut("Iterazione "+i+" controllo "+searchdeldom);
					if(parent.isSubdomainDel(searchdeldom)){
						//parent.writeOut("Trovato!");
						return searchdeldom;
					}//if  un subdom. delegato
				} catch (InvalidNameException ine){
					System.err.println("Errore in findSubdomainDel"+
						"impossibile creare nome "+domain+" e "+subs.getPrefix(i));
				}//try
			}//for
		}//if
		//parent.writeOut("Non trovato!");
		return null;
	}//findSubdomainDel
	
	//procedura che dato un nome di dominio 
	//stabilisce se operare o meno una ricerca iterativa verso l'alto
	protected boolean iterUpSearch(Domain query){
		//affinch convenga scegliere la ricerca iterativa
		//verifico che ci sia un dominio genitore comune tra 
		//i due domini e controllo la distanza dell dominio trovato 
		// da root e dom attuale
		Domain domain=parent.getDomain();
		//cerco la radice comune
		int i=1;
		while ((!(query.size()<i)) && domain.startsWith(query.getPrefix(i))){
			//parent.writeOut("i="+i+" "+domain+" inizia con "+query.getPrefix(i)+".");
			i++;
		}
		//trovata la radice comune (nel caso peggiore la root) misuro la distanza dalla root e quella dal dominio
		try{
			Domain dom=(Domain)query.getPrefix(i-1);
			//parent.writeOut("ITERUPSEARCH trovata radice comune: "+dom);
			//parent.writeOut("rispondo "+(((dom.size() -1)>=(domain.size()-dom.size()))?true:false));
			//parent.writeOut("poich dist. dalla root="+ (dom.size() -1)+
			//	"e dist. dal dominio="+(domain.size()-dom.size()));
			if ((dom.size() -1)>=(domain.size()-dom.size()))
				return true;
			else 
				return false;
		}catch (Exception e){
			System.err.println(tname+": Errore in iterUpsearch");
			return false;
		}
	}//iterUpSearch
	
	
	protected boolean sendMessage(PacketMsg msg,InetAddress addr,int port){
		try{
				//creazione del pacchetto da un messaggio
				DatagramPacket packet=msg.writeToPacket(addr,port);
				//apertura di una socket datagram su una porta qualsiasi disponibile
				DatagramSocket sock=new DatagramSocket();
				//invio del pacchetto
				sock.send(packet);
				//chiusura della socket
				sock.close();
				return true;
			} catch (Exception e) {
				System.err.println("Impossibile inviare risposta a"+senderaddr.getHostName()+
					"per la richiesta "+query.toString()+": "+ e+":"+e.getMessage());
				return false;
			}	
	}//sendMessage
	
	protected Value selectAddr(Value[] val){
		//semplicemente va in ordine...
		selectaddr=((selectaddr)%(val.length));
		return val[selectaddr++];
	}//selectAddr
	
}//IterQueryResolver
