// Name Server generico
package Dns;

import java.net.*;
import javax.naming.InvalidNameException;
import java.io.*;
import Dns.Msg.*;
import Msg.PacketMsg;

//la classe NON  astratta poich utilizzabile per servizi di naming senza replicazione
//Per una utilizzazione pratica una classe deve ereditare da name server e inizializzare la tabella

public class NameServer {
	////COSTANTI/////////////////////////////
	//Porta del servizio
	public static final int DnsPort=5656;
	/////////////////////////////
	public int servPort=5656;////
	/////////////////////////////
	//Durata time_out
	public static final int timeOut=10;


	
	////FIELDS//////////////////////////////
	//La tabella delle associazioni;
	protected Table table;
	//TimeToLive delle informazioni riguardanti questio NS;
	protected int TTL; 
	//Le informazioni di caching sono mantenute distinte;
	protected Cache cache;
	//Dominio;
	protected Domain dom;
	//Dominio di livello superiore
	protected Domain pardom;
	//timestamp unico per tutti i thread: assicura l'unicit dei messaggi
	protected int timestamp;
	
	protected NameServer(){
		dom=null;
		pardom=null;
		table=null;
	}
	
	//costruttore per root
	public NameServer (Table table,int TTL)throws Exception{
		this.dom=Domain.getRootName();
		this.pardom=null;
		this.table=table;
		this.TTL=TTL;
		this.cache=new Cache();
	}//costruttore
	
	//costruttore con tabella 
	public NameServer(Domain dom,Table table,int TTL,Domain pardom){
		this.dom=dom;
		this.pardom=pardom;
		this.table=table;
		this.TTL=TTL;
		this.cache=new Cache();
	}//costruttore
	
	//costruttore con tabella vuota 
	public NameServer(Domain dom,int TTL,Domain pardom){
		this.dom=dom;
		this.pardom=pardom;
		this.table=new Table();
		this.TTL=TTL;
		this.cache=new Cache();
	}//costruttore
	
	//Metodo principale
	//il Name Server si mette in ascolto:
	public void queryResolution(){
		DatagramSocket listsock=null;
		try{
			listsock=new DatagramSocket(servPort);
		} catch (SocketException e){
			System.err.println("Errore in apertura socket:"+e.getMessage());
			System.exit(1);
		}
		try{
			writeOut("Server "+dom+" attivo su "+InetAddress.getLocalHost()+" "+servPort);
		}catch(UnknownHostException uhe){
			writeOut("Impossibile determinare nome dell'host");
		}
		while(true){
			try {
				DatagramPacket packet=new DatagramPacket(new byte[PacketMsg.dimPacket],PacketMsg.dimPacket);
				listsock.receive(packet);
				QueryMsg msg=(QueryMsg)PacketMsg.readFromPacket(packet);
				writeOut("*****************************************************************\n"
				+"** Ricevuto messaggio di richiesta "+msg);
				IterQueryResolver thread=new IterQueryResolver(msg,packet.getAddress(),packet.getPort(),this); 
				writeOut("** Attivo thread "+thread.getName());
				thread.start();	
			} catch (ClassCastException cce){
				writeOut("** Ricevuto Messaggio Sconosciuto:"+cce.getMessage());
			} catch (StreamCorruptedException sce){
				writeOut("** Errore nella lettura del messaggio ricevuto:"+sce);
			}catch (Exception e){
				writeOut("** Errore in ricezione richiesta:"+e);
			} finally{
				writeOut("*****************************************************************");
			}//try
		}//while
	}//queryResolution
	
	//controllo se un dominio  un sottodominio esistente
	//di quello del NS
	public boolean isSubdomain(Domain d){
	 return table.containsSubdomain(d);
	}//isSubdomain
	
	//controllo se un dominio  un sottodominio esistente
	//delegato di quello del NS
	public boolean isSubdomainDel(Domain d){
		return table.containsSubdomainDel(d);
	}//isSubdomainDel
	
	public Domain getDomain(){
		//necessario per evitare che eventuali effetti collaterali
		//cambino il dominio
		return (Domain)dom.clone();
	}//getDomain
	
	public Domain getParentDomain(){
		return (Domain)pardom.clone();
	}//getDomain
	
	public int getTimeToLive(){ 
		return TTL;
	}
	
	//metodi di accesso  alla tabella 	
	public Value[] lookup(DNSName name) throws BindNotFoundException{
		return table.get(name);
	}//lookup
	
	//metodi di accesso alla cache 
  public DNSName cacheSearch(DNSName query){
		//ricerco il nome richiesto
		DNSName found=null;
		Value[] res=cache.get(query);
		if(res!=null){
			try{
				found=new DNSName(query.getType(),query.getDomain());
			}catch(InvalidNameException ine){
				System.err.println("Fatal error"+ine.getMessage());
				System.exit(1);
			}
			return found;
		} else{
			//devo ricercare una "scorciatoia"
			//parto dal dominio della query cerco finch non trovo in cache
			//non trovato se: arrivo alla root
			//mi fermo se arrivo ad un dominio parente del dom attuale(compreso)
			Domain domquery=query.getDomain();
			found=null;
			try{
				do{
					//cerco il NS corrispondente al dominio
					found=new NSName(domquery);
					res=cache.get(found); 
					if(res==null)domquery=(Domain)domquery.getPrefix(domquery.size()-1);
					//writeOut"nuovo domquery="+domquery);
					//writeOut"esco dal ciclo?="+ !((res==null)&&(!(domquery.equals(Domain.getRootName())))&&(!(dom.startsWith(domquery)))));
				}while((res==null)&&(!(domquery.equals(Domain.getRootName())))&&(!(dom.startsWith(domquery))));
		
				if (res==null){
					found=null;
					//writeOut"CACHE: trovato niente :(");
				}else{
					//writeOut"CACHE: trovato refer"+found);
				}				
			} catch(Exception e){
				System.err.println("Errore "+e);
				return null;
			}
		}
		return found;
	}
	
	public Value[] cacheLookup(DNSName query){
		//ricerco il nome richiesto
		return cache.get(query);
	}
	
	public void cache(Binding b,int TTL){
		//inserisco il nome richiesto
		cache.put(b,TTL);
	}
	
	public int getCacheTimeToLive(DNSName query){
		return cache.getTimeToLive(query);
	}
	
	//metodo che i thread devono richiamare per ottenere
	//un timestamp con cui identificare univocamente i messaggi:
	public synchronized int stamp(){
		if (++timestamp==Integer.MAX_VALUE)timestamp=0;
		return timestamp;
	}
	
	protected void writeOut(String s){
		System.out.println(s);
	}
	
}//class NameServer




