package Dns;

import java.io.*;
import java.net.*;
import iaik.security.rsa.RSASignature;
import iaik.security.rsa.*;
import iaik.java.security.*;
import com.entrust.util.*;
import com.entrust.security.provider.RSA;
import iaik.security.rsa.ShaRSASignature;
import java.util.*;
import Security.RSAKeyGen;
import Dns.Gui.*;

public class PrimaryServer extends NameServer{
	public static final String datafile="nsdata.ini";
	//porta di ascolto per l'update dei secondari
	protected final static int UpdatePort=DnsPort;
	protected File dfile;
	//contatore degli aggiornamenti:
	protected long tablever=0;
	
	PrimaryGui gui;

	//costruttore
	public PrimaryServer() throws Exception{
		this(datafile);
	}
	
	
	public PrimaryServer(String file) throws Exception{
		dfile=new File(file);
		if (!dfile.exists()){
			//prima esecuzione, crea il file:
			init(dfile);
		}
		
		//lettura dei dati da file
		ObjectInputStream ois=new ObjectInputStream(new FileInputStream(dfile));
		System.out.println("Inizializzazione del server...");
		servPort=ois.readInt();
		System.out.println("(*) Porta di ascolto: "+servPort);
		dom=(Domain)ois.readObject();
		System.out.println("(*) Dominio: "+dom);
		if(!dom.equals(Domain.getRootName())){
			pardom=(Domain)ois.readObject();
			System.out.println("(*) Dominio superiore: "+pardom);
		} else pardom=null;
		table=(Table)ois.readObject();
		System.out.println("(*) Tabella corrente: \n"+table);
		TTL=ois.readInt();
		System.out.println("(*) TimeToLive delle informazioni : "+TTL+" sec.");
		cache=new Cache();
		ois.close();
		System.out.println("Inizializzazione terminata.");	
		
		//Thread di ascolto per l'update dei secondari:
		new UpdateThread(this).start();
		
		//interfaccia grafica
		gui=new PrimaryGui(this);
		gui.show();
		
		//attivazione server
		queryResolution();

	}//costruttore
	
	//inizializzazione
	protected void init(File file){
		//PrimaryInitGui frame=new PrimaryInitGui(datafile);
		PrimaryInitGui frame=new PrimaryInitGui(file);
		frame.start();
		try{
			frame.join();
		}catch(InterruptedException e){
			System.out.println(e);
		}
	}

	public Table update(){
		return table.makeCopy();
	}

	public void bind(Binding bind) throws Exception{
		DNSName name=bind.getName();
		
		Domain domain=name.getDomain();
		
		if(name.getType() instanceof UserType){
		//inserimento utente (solo se appartenente a dominio o sottodominio non del.)
		if(!((domain.equals(getDomain()))||(isSubdomain(domain)&&(!isSubdomainDel(domain)))))
			throw new Exception("L'utente deve appartenere al dominio o ad un sottodominio non delegato.");
		} else if (name.getType() instanceof NSType){
			//inserimento NS: sono ammessi solo root o sottodomini delegati o dominio superiore:
			if(!((domain.equals(Domain.getRootName()))||(isSubdomainDel(domain)) || (domain.equals(getParentDomain()))))
				throw new Exception("Il NS deve essere di root o del dom. superiore o di un sottodominio delegato.");
		} else if (name.getType() instanceof ASType){
			//inserimento AS: ammesso solo per il dominio corrente:
			if(!(domain.equals(getDomain())))
				throw new Exception("L'AS deve appartenere al dominio.");
		
		}
		table.put(bind);
		//registrazione su file:
		fileWrite();
	}
	
	public void unbind(DNSName name) throws Exception{
		Domain dom=name.getDomain();
		if((!dom.equals(getParentDomain())) && (!dom.equals(Domain.getRootName()))){
			table.remove(name);
			//registrazione su file:
			fileWrite();
		} else throw new Exception("Impossibile eliminare dati.");
	}
	
	public void addSubdomain(Domain dom) throws Exception{
		//il dominio deve essere un sottodominio di questo o
		//di un sottodominio gi esistente non delegato:
		Domain tocheck=(Domain)dom.getPrefix(dom.size() -1);
		if(tocheck.equals(getDomain()) || (isSubdomain(tocheck) && !isSubdomainDel(tocheck))){
			table.addSubdomain(dom);
			writeOut("Aggiunto sottodominio "+dom +".");
		}else
			throw new Exception("il dominio deve essere un sottodominio di questo o di uno gi esistente non delegato.");
		fileWrite();
	}
	
	public void delSubdomain(Domain dom) throws Exception{
		//il dominio deve essere un sottodominio non delegato: 
		
		if(isSubdomain(dom) && !isSubdomainDel(dom)){
			table.delegateSubdomain(dom);
			writeOut("Sottodominio "+dom +" delegato.");
		 } else
			throw new Exception("il dominio deve essere un sottodominio non delegato.");
		fileWrite();
	}
	
	protected synchronized void fileWrite() throws Exception{
		try{
			dfile.delete();
			dfile.createNewFile();
			ObjectOutputStream oos=new ObjectOutputStream(new FileOutputStream(dfile));
			oos.writeInt(servPort);
			oos.writeObject(dom);
			if(!dom.equals(Domain.getRootName())){
				oos.writeObject(pardom);
			} 
			oos.writeObject(table);
			oos.writeInt(TTL);			
			oos.close();
			tablever++;
		}catch(IOException ioe){
			System.err.println("Errore in scrittura file dati");
		}
	}
	
	protected void writeOut(String s){
		gui.writeOut(s);
	}
	
	public void stop(){
		System.out.println("Chiusura...");
		System.exit(0);
	}
	
	public static void main(String args[]){
		
		try{
			if (args.length==0){
				PrimaryServer server=new PrimaryServer();
			}else if (args.length==1){
				PrimaryServer server=new PrimaryServer(args[0]);
			}else 
				System.out.println("Usage: java Dns.PrimaryServer [LogFile]");
		}catch(Exception e){
			System.out.println("Errore: "+e);
			System.exit(1);
		}
	}
}//PrimaryServer

//////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////
class UpdateThread extends Thread{

	PrimaryServer server;
	
	public UpdateThread(PrimaryServer server){
		this.server=server;
	}//costruttore

	public void run(){
		Socket clientsock=null;//socket connessione
		ServerSocket listeningsock=null; //socket di ascolto
		try{
			//Crea la socket di ascolto:
			//listeningsock = new ServerSocket(server.UpdatePort);
			listeningsock = new ServerSocket(server.servPort);
		} catch (Exception e){
			System.err.println("Errore nella socket di ascolto per l'update:"+e.getMessage());
			System.exit(1);
		}
		while(true){
			try{
				//Crea la Socket di connessione
			  clientsock=listeningsock.accept();
			} catch(IOException e){
				System.err.println("Errore nella socket di connessione per l'update:"+e.getMessage());
				System.exit(1);
			}
			//Gestione del servizio:
			(new UpdateSecondaryThread(clientsock,server)).start();
		}//while
	}//run()

}//UpdateThread

class UpdateSecondaryThread extends Thread{
	//Chiavi
	protected final static RSAPrivateKey privkey=RSAKeyGen.readPrivate("priv.privkey");
	protected final static RSAPublicKey pubkey=RSAKeyGen.readPublic("pub.pubkey");
	
	protected Socket socket;
	protected PrimaryServer server;
	
	public UpdateSecondaryThread(Socket socket, PrimaryServer server){
		this.socket=socket;
		this.server=server;
	}

	public void run(){
		try{
			ObjectOutputStream sockoos=new ObjectOutputStream(socket.getOutputStream());
			ObjectInputStream sockois=new ObjectInputStream(socket.getInputStream());
			long secver=sockois.readLong();
			//ois.close();
			long prver=server.tablever;
			System.out.println("Letta vers. "+secver+" versione corrente "+prver);
			sockoos.writeLong(prver);
			sockoos.flush();
			if(prver==secver){	
				System.out.println("Aggiornamento non necessario.");
				//oos.close();
				//socket.close();
			} else {
				System.out.println("Inizio aggiornamento.");
				//trasformazione della tabella in un array di bytes
				ByteArrayOutputStream baos=new ByteArrayOutputStream();
				ObjectOutputStream oos=new ObjectOutputStream(baos); 
		 		oos.writeObject(server.getDomain());
				if(!server.getDomain().equals(Domain.getRootName())){
					oos.writeObject(server.getParentDomain());
				}
				oos.writeObject(server.update());
				oos.writeInt(server.getTimeToLive());
				oos.close();
				byte[] data =baos.toByteArray();
				baos.close();
				
				//creazione signature:
				Util.initCiphers();
				ShaRSASignature sign;
	
				byte[][] signature;
				
				int length;
				
				if(data.length%512!=0)
					length=(data.length/512)+1;
				else 
					length=(data.length/512);
				//System.out.println("Lungh. dati = "+data.length+" ,steps= "+length);
	
				signature=new byte [length][64];
				int i=0;
				byte[] datatosign;
				
				while(i<length){
					if(i==(length-1)) datatosign=new byte[data.length-((length-1)*512)];
						else datatosign=new byte[512];
					for(int j=0;(j<512 && (i*512+j)<data.length);j++){
						datatosign[j]=data[i*512+j];
					}
					//Signature
					sign=new ShaRSASignature();
					sign.initSign(privkey);
					//System.out.println("Lungh. dati "+i+" ==>"+datatosign.length);
					sign.update(datatosign);
					signature[i++]=sign.sign();
				}
				
				//invio della tabella e signature sulla socket
				sockoos.writeObject(data);
				sockoos.writeObject(signature);
				socket.close();
			}//else
		}catch(Exception e){
			System.err.println("Errore "+e);
		}
	}
}
