package Dns;

import Exceptions.*;
import java.io.*;
import java.net.*;
import iaik.security.rsa.RSASignature;
import iaik.security.rsa.*;
import iaik.java.security.*;
import com.entrust.util.*;
import com.entrust.security.provider.RSA;
import iaik.security.rsa.ShaRSASignature;
import Security.RSAKeyGen;
import java.awt.event.*;
import javax.swing.Timer;

public class SecondaryServer extends NameServer implements ActionListener{
	//Chiavi
	protected final static RSAPrivateKey privkey=RSAKeyGen.readPrivate("priv.privkey");
	protected final static RSAPublicKey pubkey=RSAKeyGen.readPublic("pub.pubkey");
	
	//secondi da attendere prima della richiesta di update:
	protected int checktime;
	//dati del NS primario
	InetAddress primaryaddr;
	int primaryport;
	//Timer per timout
	Timer timeout;

	//Versione dell'aggiornamento
	long ver=-1;
	
	public SecondaryServer(InetAddress primaryaddr,int primaryport,int checktime,int myport){
		this.servPort=myport;
		this.primaryaddr=primaryaddr;
		this.primaryport=primaryport;
		this.checktime=checktime;
		//connessione:
		try{
			if(requestUpdate()){
				//timeout:
				timeout=new Timer(checktime*1000,this);
				timeout.start();
				//attivazione NS:
				queryResolution();
			} else {
				System.err.println("Primo aggiornamento necessario.");
				System.exit(1);
			}
		}catch (SecurException e){
			System.err.println("Errore di sicurezza durante il primo aggiornamento.");
			System.exit(1);
		}
	}//costruttore
	
	public SecondaryServer(InetAddress primaryaddr,int checktime){
		this(primaryaddr,DnsPort,checktime,DnsPort);
	}

	public void actionPerformed(ActionEvent e){
		try{
			requestUpdate();	
		} catch (SecurException se){
			System.err.println(e);
		}
	}

	protected boolean requestUpdate() throws SecurException{
		try{
			Socket s=new Socket(primaryaddr,primaryport);
			System.out.println("Socket aperta");
			ObjectOutputStream sockoos=new ObjectOutputStream(s.getOutputStream());
			System.out.println("Output Stream aperta");
			
			sockoos.writeLong(ver);
			sockoos.flush();
			System.out.println("Inviata versione tabella "+ver);
			//oos.close();
			ObjectInputStream sockois=new ObjectInputStream(s.getInputStream());
			System.out.println("Input Stream aperta");
			long prver=sockois.readLong();
			System.out.println("Ricevuta versione tabella "+prver);
			if(ver!=prver){
				System.out.println("Procedo all'aggiornamento...");
				byte[] data=(byte[])sockois.readObject();
				System.out.println("Ricevuti dati. "+data.length);
				byte[][] signature=(byte[][])sockois.readObject();
				System.out.println("Ricevuta signature "+signature.length);
				sockois.close();
			
				//controllo signature:
				Util.initCiphers();
				ShaRSASignature sign;
				
				for(int i=0;i<signature.length;i++){
					byte[] tocheck;
					if(i==(signature.length-1)) tocheck=new byte[data.length-((signature.length-1)*512)];
					else tocheck=new byte[512];;	
					for(int j=0;(j<512 && (i*512+j)<data.length);j++){
						tocheck[j]=data[(i*512+j)];
					}		
					sign=new ShaRSASignature();
					sign.initVerify(pubkey);
					sign.update(tocheck);
	 				if (sign.verify(signature[i])) {
						System.out.println("Signature "+i+" correct!");
					}else{
						System.out.println("Signature "+i+" *NOT* correct!");
						throw new SecurException("Signature "+i+" *NOT* correct!");
					}
				}//for
		
				//ricostuzione tabella:
				ByteArrayInputStream bais=new ByteArrayInputStream(data);
	
				System.out.println("inizio ois");	
				ObjectInputStream ois=new ObjectInputStream(bais);
				dom=(Domain)ois.readObject();
				System.out.println("letto dom: "+dom);
				if(!dom.equals(Domain.getRootName())){
					pardom=(Domain)ois.readObject();
					System.out.println("letto pardom: "+pardom);
				}
				table=(Table)ois.readObject();
				System.out.println("letta table: "+table);
				TTL=ois.readInt();
				System.out.println("letto TTL: "+TTL);
				//aggiornamento versione:
				ver=prver;	
			} else {
				System.out.println("Tabella gi aggiornata.");
				sockois.close();
				s.close();
			
			}
			return true;
		} catch (ConnectException e){
			System.err.println("Server primario non presente:"+e);
			return false;
			//System.exit(1);
		} catch (Exception e){
			System.err.println("Errore"+e);
			System.exit(1);
			return false;
		}
	}//requestUpdate
	
	public static void main(String args[]){
		if (args.length!=4){
				System.out.println("Usage:java SecondaryServer host port seconds myport");
				System.exit(1);
			}
		try{
			SecondaryServer server=new SecondaryServer(InetAddress.getByName(args[0]),Integer.parseInt(args[1]),Integer.parseInt(args[2]),Integer.parseInt(args[3]));
		}catch(Exception e){
			System.out.println("Usage:java SecondaryServer host port seconds :\n"+ e);
		}
	}

}//SecondaryServer
