package Security;

import javax.crypto.KeyGenerator;
import iaik.security.cipher.*;
import javax.crypto.Cipher;
import iaik.java.security.Security;
import iaik.java.security.SecureRandom;
import com.entrust.security.provider.RSA; 
import com.entrust.util.*;
import iaik.security.rsa.*;
import iaik.java.security.*;
import java.io.*;


public class RSAKeyGen {
	RSAPublicKey rsa_pub_key=null;
	RSAPrivateKey rsa_priv_key=null;
	
	//Costruttore senza argomenti: genera chiave pubblica e privata
	public RSAKeyGen() throws Exception{
		// inizializzazione necessaria!: altrimenti eccezione NoSuchAlgorithmException(): RSA algorithm not initialized
		Util.initCiphers();
		/* KeyPairGenerator con algoritmo RSA*/
		RSAKeyPairGenerator key_gen;
		//key_gen = KeyPairGenerator.getInstance("RSA");
		key_gen = new RSAKeyPairGenerator();
		/* chiave a 512 bit (default 1024)*/
		key_gen.initialize(512);
		/*generazione delle chiavi*/
		java.security.KeyPair key_pair = key_gen.generateKeyPair();
		/*la chiave pubblica*/
		rsa_pub_key = (RSAPublicKey)key_pair.getPublic();
		//System.out.println("c.pubblica="+rsa_pub_key+"\n");
		/*la chiave pprivata*/
		rsa_priv_key = (RSAPrivateKey)key_pair.getPrivate();
		//System.out.println("c.segreta="+rsa_priv_key+"\n");
	}
	
	//Costruttore con argomento: legge chiavi da  file
	public RSAKeyGen(String privfile,String pubfile) throws Exception{
		FileInputStream istream = new FileInputStream(privfile);
		ObjectInputStream p = new ObjectInputStream(istream);
		rsa_priv_key=(RSAPrivateKey)p.readObject();
		istream.close();
		p.close();
		istream = new FileInputStream(pubfile);
		p = new ObjectInputStream(istream);
		rsa_pub_key=(RSAPublicKey)p.readObject();
	}
	
	public RSAPublicKey getPublic(){
		return 	rsa_pub_key;
	}
	
	public RSAPrivateKey getPrivate(){
		return 	rsa_priv_key;
	}
	
	public boolean writePublic(String nomefile){
		try{
			FileOutputStream fostream = new FileOutputStream(nomefile);
			ObjectOutputStream ostream = new ObjectOutputStream(fostream);
			ostream.writeObject(rsa_pub_key);
			ostream.close();
			fostream.close();
			return true;
		} catch (Exception e){
			System.err.println(e.getMessage());
			return false;
		} 
	}
	
	public boolean writePrivate(String nomefile){
		try{
			FileOutputStream fostream = new FileOutputStream(nomefile);
			ObjectOutputStream ostream = new ObjectOutputStream(fostream);
			ostream.writeObject(rsa_priv_key);
			return true;
		} catch (Exception e){
			System.err.println(e.getMessage());
			return false;
		} 
	}
	
	public static RSAPublicKey readPublic(String nomefile) {
		try{
			FileInputStream istream = new FileInputStream(nomefile);
			ObjectInputStream p = new ObjectInputStream(istream);
			RSAPublicKey key=(RSAPublicKey)p.readObject();
			istream.close();
			p.close();
			return key;
		}catch(Exception e){
			System.err.println("Errore in lettura della chiave "+nomefile);
			return null;
		}
	} 
	
	public static RSAPrivateKey readPrivate(String nomefile) {
		try{
			FileInputStream istream = new FileInputStream(nomefile);
			ObjectInputStream p = new ObjectInputStream(istream);
			RSAPrivateKey key=(RSAPrivateKey)p.readObject();
			istream.close();
			p.close();
			return key;
		}catch(Exception e){
			System.err.println("Errore in lettura della chiave "+nomefile);
			return null;
		}
	}
	
	public static void main(String[] args){
	try{	
		RSAKeyGen kg=new RSAKeyGen();
		//System.out.println("c.pubblica="+kg.getPublic()+"\n");
		kg.writePublic("pub.pubkey");
		kg.writePrivate("priv.privkey");
		//kg=new RSAKeyGen("c:/documenti/progettordc/priv.txt","c:/documenti/progettordc/pub.txt");
		//System.out.println("c.pubblica="+kg.getPublic()+"\n");
	} catch (Exception e){e.printStackTrace();}
	}
}
